import 'dart:async';

import 'package:flex/flex.dart';
import 'package:plexus/plexus.dart';

/// Represents a class holding a message
abstract class IHasShortMessage {
  /// The short message for this event. Should be less than 40 characters.
  String get message;
}

/// Represents a class holding a timestamp
abstract class IHasTimestamp {
  /// The timestamp for this event.
  DateTime get timestamp;
}

/// Represents a class holding a long message
abstract class IHasLongMessage {
  /// The long message for this event. Should contain the same information as [IHasMessage.message] and then expand on it.
  @nullable
  String get longMessage;
}

/// Represents a class holding a source
abstract class IHasSource {
  /// The source of this event.
  Module get source;
}

/// Represents a class holding a blob of type [B]
abstract class IHasBlob<B> {
  /// Non-string message passed to Logger.
  @nullable
  B get blob;
}

/// Represents a class holding a level
abstract class IHasLevel {
  /// The level of this event.
  Level get level;
}

/// Represents a class holding an error and optional stacktrace
abstract class IHasError {
  /// Associated error (if any) when recording errors messages.
  @nullable
  Object get error;

  /// Associated stackTrace (if any) when recording errors messages.
  @nullable
  StackTrace get stackTrace;
}

/// Represents a class holding a zone
abstract class IHasZone {
  /// Zone of the calling code which resulted in this LogRecord.
  Zone get zone;
}

/// Represents a class holding a state of init or not init
abstract class IHasIsInit {
  /// Does this event take place during initialization
  bool get init;
}

/// An event which will be logged by a [Logger]'s individual handlers.
///
/// [source] will usually be set by a Logger.
/// Manually setting [timestamp] is discouraged.
abstract class Event<B> implements IHasSource, IHasShortMessage, IHasBlob<B>, IHasTimestamp, IHasLevel, IHasError, IHasZone, IHasIsInit {
  /// Creates a new instance of [Event]
  Event(@nullable DateTime timestamp) : timestamp = timestamp ?? DateTime.now();

  static StdEvent<String> parse(String encoded, Module source, @nullable Zone zone) {
    assert(encoded.startsWith(encodedEventHeader));

    final e = encoded.split(encodedEventSeparator);

    return StdEvent<String>(source, Level.encodedToPlexus[e[2]], e[3], blob: e.length == 5 ? e[4] : null);
  }

  @override
  final DateTime timestamp;
}

/// The default implementation of [Event]
class StdEvent<B> extends Event<B> implements IHasLongMessage {
  /// Creates a new instance of [StdEvent]
  StdEvent(this.source, this.level, this.message, {this.longMessage, this.error, this.stackTrace, this.blob, @nullable Zone zone, this.init = false, @nullable DateTime timestamp})
      : zone = zone ?? Zone.current,
        super(timestamp);

  @override
  final Module source;

  @override
  final Level level;

  @override
  final String message;

  @override
  @nullable
  final String longMessage;

  /// Non-string message passed to Logger.
  @override
  @nullable
  final B blob;

  /// Associated error (if any) when recording errors messages.
  @override
  @nullable
  final Object error;

  /// Associated stackTrace (if any) when recording errors messages.
  @override
  @nullable
  final StackTrace stackTrace;

  /// Zone of the calling code which resulted in this LogRecord.
  @override
  final Zone zone;

  /// Does this event take place during initialization
  @override
  final bool init;
}

/// An implementation of [Event] representing a change in authentication state
class AuthEvent<B> extends Event<B> implements IHasLongMessage {
  /// Creates a new instance of [AuthEvent]
  AuthEvent(this.source, this.level, this.message, this.signal, {this.longMessage, this.error, this.stackTrace, this.blob, @nullable Zone zone, this.init = false, @nullable DateTime timestamp})
      : zone = zone ?? Zone.current,
        super(timestamp);

  @override
  final Module source;

  @override
  final Level level;

  @override
  final String message;

  @override
  @nullable
  final String longMessage;

  /// Non-string message passed to Logger.
  @override
  @nullable
  final B blob;

  /// Associated error (if any) when recording errors messages.
  @override
  @nullable
  final Object error;

  /// Associated stackTrace (if any) when recording errors messages.
  @override
  @nullable
  final StackTrace stackTrace;

  /// Zone of the calling code which resulted in this LogRecord.
  @override
  final Zone zone;

  /// Does this event take place during initialization
  @override
  final bool init;

  /// The [AuthSignal] of this event.
  final AuthSignal signal;
}

/// Represents a test used to validate user input given [context] representing past input and [input].
typedef bool InputTest(List<String> context, String input);

/// An implementation of [Event] representing a request for user input
class RequestInputEvent<B> extends Event<B> {
  /// Creates a new instance of [RequestInputEvent]
  RequestInputEvent(this.source, this.context, this.message, this.test, {this.error, this.stackTrace, this.blob, @nullable Zone zone, this.init = false, @nullable DateTime timestamp})
      : zone = zone ?? Zone.current,
        super(timestamp);

  @override
  final Module source;

  @override
  final Level level = Level.PRINT;

  /// This will be used for the input prompt. It should usually contain the context.
  @override
  final String message;

  /// Non-string message passed to Logger.
  @override
  @nullable
  final B blob;

  /// Associated error (if any) when recording errors messages.
  @override
  @nullable
  final Object error;

  /// Associated stackTrace (if any) when recording errors messages.
  @override
  @nullable
  final StackTrace stackTrace;

  /// Zone of the calling code which resulted in this LogRecord.
  @override
  final Zone zone;

  /// Does this event take place during initialization
  @override
  final bool init;

  /// Contains all commands/arguments parsed prior to this.
  final List<String> context;

  /// Validates [current] input with access to [context] previous (valid) inputs.
  final InputTest test;

  /// Validates the user's input.
  bool testInput(String current) => test(context, current);
}
