import 'package:dart_numerics/dart_numerics.dart';
import 'package:logging/logging.dart' as dlog;
import 'package:plexus/plexus.dart';

/// [Level]s to control logging output. Logging can be enabled to include all
/// levels above certain [Level]. [Level]s are ordered using an integer
/// value [Level.value]. The predefined [Level] constants below are sorted as
/// follows (in descending order): [Level.SHOUT], [Level.SEVERE],
/// [Level.WARNING], [Level.INFO], [Level.CONFIG], [Level.FINE], [Level.FINER],
/// [Level.FINEST], and [Level.ALL].
///
/// We recommend using one of the predefined logging levels. If you define your
/// own level, make sure you use a value between those used in [Level.ALL] and
/// [Level.OFF].
class Level implements Comparable<Level> {
  const Level._(this.name, this.value, this.ansiCode, this.dlogLevel);

  /// This level's name
  final String name;

  /// This level's name wrapped with [ansiCode]
  String get wrappedName => ansiCode.wrap(name);

  /// Unique value for this level. Used to order levels, so filtering can
  /// exclude messages whose level is under certain value.
  final int value;

  /// The [AnsiCode] used for formatting in the console.
  final AnsiCode ansiCode;

  /// The corresponding Dart Log level for this Plexus Log level
  final dlog.Level dlogLevel;

  @override
  bool operator ==(dynamic other) => other is Level && value == other.value && name == other.name;

  /// Whether this [Level] is less than [other]
  bool operator <(Level other) => value < other.value;

  /// Whether this [Level] is less than or equal to [other]
  bool operator <=(Level other) => value <= other.value;

  /// Whether this [Level] is greater than [other]
  bool operator >(Level other) => value > other.value;

  /// Whether this [Level] is greater than or equal to [other]
  bool operator >=(Level other) => value >= other.value;

  @override
  int compareTo(Level other) => value - other.value;

  @override
  int get hashCode => value;

  @override
  String toString() => name;

  /// Special key to turn on logging for all levels ([value] = 0).
  static const Level ALL = Level._('ALL', 0, AnsiCode.none, dlog.Level.ALL);

  /// Special key to turn off all logging ([value] = 2000).
  static const Level DISABLED = Level._('OFF', int64MaxValue, AnsiCode.none, dlog.Level.OFF);

  /// Key for highly detailed tracing ([value] = 300).
  static const Level FINEST = Level._('DEBUG/3', 300, AnsiCode.lightBlue, dlog.Level.FINEST);

  /// Key for fairly detailed tracing ([value] = 400).
  static const Level FINER = Level._('DEBUG/2', 400, AnsiCode.lightBlue, dlog.Level.FINER);

  /// Key for tracing information ([value] = 500).
  static const Level FINE = Level._('DEBUG/1', 500, AnsiCode.lightBlue, dlog.Level.FINE);

  /// Key for static configuration messages ([value] = 600).
  static const Level CONFIG = Level._('CONFIG', 600, AnsiCode.lightYellow, dlog.Level.CONFIG);

  /// Key for informational messages ([value] = 700).
  static const Level INFO = Level._('INFO', 700, AnsiCode.lightBlue, dlog.Level.INFO);

  /// Key for success messages ([value] = 800).
  static const Level SUCCESS = Level._('SUCCESS', 800, AnsiCode.lightGreen, dlog.Level.INFO);

  /// Key for messages that would otherwise be output by print() ([value] = 800).
  static const Level PRINT = Level._('PRINT', 800, AnsiCode.none, dlog.Level.INFO);

  /// Key for potential problems ([value] = 800).
  static const Level WARNING = Level._('WARNING', 800, AnsiCode.lightYellow, dlog.Level.WARNING);

  /// Key for errors ([value] = 900).
  static const Level ERROR = Level._('ERROR', 900, AnsiCode.orange, dlog.Level.SEVERE);

  /// Key for fatal errors that will crash the program ([value] = 1000).
  static const Level FATAL = Level._('FATAL', 1000, AnsiCode.lightRed, dlog.Level.SEVERE);

  /// Key for extra debugging loudness ([value] = 1200).
  static const Level SHOUT = Level._('SHOUT', 1200, AnsiCode.lightYellow, dlog.Level.SHOUT);

  /// A list of all Plexus Log levels
  static const List<Level> LEVELS = [ALL, FINEST, FINER, FINE, CONFIG, INFO, SUCCESS, WARNING, ERROR, FATAL, SHOUT, DISABLED];

  /// A map of Dart Log levels to Plexus Log levels
  static Map<dlog.Level, Level> dlogToPlexus = {
    dlog.Level.ALL: ALL,
    dlog.Level.OFF: DISABLED,
    dlog.Level.FINEST: FINEST,
    dlog.Level.FINER: FINER,
    dlog.Level.FINE: FINE,
    dlog.Level.CONFIG: CONFIG,
    dlog.Level.INFO: INFO,
    dlog.Level.WARNING: WARNING,
    dlog.Level.SEVERE: ERROR,
    dlog.Level.SHOUT: SHOUT,
  };

  static Map<String, Level> encodedToPlexus = {
    'Level.ALL': ALL,
    'Level.FINEST': FINEST,
    'Level.FINER': FINER,
    'Level.FINE': FINE,
    'Level.CONFIG': CONFIG,
    'Level.INFO': INFO,
    'Level.SUCCESS': SUCCESS,
    'Level.WARNING': WARNING,
    'Level.ERROR': ERROR,
    'Level.FATAL': FATAL,
    'Level.SHOUT': SHOUT,
    'Level.DISABLED': DISABLED,
  };
}
