import 'dart:async';
import 'dart:collection' show Queue;

import 'package:collection/collection.dart' show ListEquality;
import 'package:dartx/dartx.dart';
import 'package:meta/meta.dart';
import 'package:flex/expose.dart';
import 'package:flex/flex.dart';
import 'package:plexus/plexus.dart';

/// A function used to generate a [Module]
typedef Module ModuleBuilder(List<String> package, Level level);

/// A function that handles [Event]
typedef void EventHandler(Event event);

/// Implemented by any class that wants to support Plexus logging operations.
abstract class ILog {
  /// Automatically record stack traces for any message of this level or above.
  ///
  /// Because this is expensive, this is off by default.
  Level get recordStackTraceAtLevel;

  /// Effective level considering the levels established in this logger's
  /// parents.
  Level get level;

  /// Override the level for this particular [Module] and its children.
  set level(Level value);

  /// Whether a message for [value]'s level is loggable in this logger.
  bool isLoggable(Level value);

  /// Adds a event to be logged with [message] at [logLevel]
  void log(Level logLevel, dynamic message, {@nullable String longMessage, @nullable Object error, @nullable StackTrace stackTrace, @nullable Zone zone, bool init = false, @nullable dynamic blob});

  /// Adds a event for a change in authentication state to be logged with [message] at [logLevel]
  void logAuth(Level logLevel, dynamic message, AuthSignal signal, {@nullable String longMessage, @nullable Object error, @nullable StackTrace stackTrace, @nullable Zone zone, bool init = false, @nullable dynamic blob});

  /// Adds an event for requesting input in the console
  /// 
  /// Do not use in Flutter
  void logRequestInput(List<String> context, String prompt, InputTest test, {@nullable Object error, @nullable StackTrace stackTrace, @nullable Zone zone, bool init = false, @nullable dynamic blob});
}

/// The base module class. The default implementation is [StdModule].
abstract class Module implements ILog {
  /// Returns the module matching [package], generating a new [StdModule] if it cannot be found.
  factory Module(List<String> package, {Level level = Level.INFO, ModuleBuilder builder = generate}) => Module.getModuleSafely(package: package, level: level, builder: builder);

  /// Returns the module matching [package], generating a new [StdModule] if it cannot be found.
  factory Module.fromString(String package, {Level level = Level.INFO, ModuleBuilder builder = generate}) => Module.getModuleSafely(stringPackage: package, level: level, builder: builder);

  /// Returns the existing module matching [package], throwing an error if it cannot be found.
  factory Module.existing(List<String> package) => modules[package.join('.')];

  /// Returns the existing module matching [package], throwing an error if it cannot be found.
  factory Module.existingFromString(String package) => modules[package];

  /// This internal constructor should be used by subclasses of [Module]
  @protected
  const Module.internal();

  /// This module's internal level. **DO NOT USE THIS! Use [level] instead!**
  @protected
  Level get internalLevel;

  /// This module's internal level. **DO NOT USE THIS! Use [level] instead!**
  @protected
  set internalLevel(Level newLevel);

  /// The package of this module
  UnmodifiableListView<String> get package;

  /// This modules event handlers
  @protected
  List<String> get internalEventHandlers;

  /// This modules event handlers, including those inheritted from parents.
  @protected
  List<String> get eventHandlers;

  /// This module.
  String get immediateModule;

  /// This module's parent
  Module get parent;

  /// The name of this module
  String get name;

  /// The full name of this logger, which includes the parent's full name.
  String get fullName;

  /// This module's package in the format "some.module.child.thisone"
  String get packageifiedPackage;

  /// This module's package in the format "/some/module/child/thisone"
  String get pathifiedPackage;

  /// This module's package in the format "/some/module/child/thisone.dart"
  ///
  /// Can be useful for finding the file a log is coming from (we don't find the file for you. The file must match the package)
  String get pathifiedPackageWithExtension;

  //String get fullyQualifiedPackage => package.isNotEmpty ? '$prefix:$pathifiedPackage' : '/';

  /// Subscribes to events. You must unsubscribe using the [unsubscribe] method.
  StreamSubscription<Event> subscribe(String id, Value1Callback<Event, void> callback);

  /// Unsubscribes handler [id]
  Future<void> unsubscribe(String id);

  /// Unsubscribes all handlers on this module and all children
  Future<void> unsubscribeAll();

  /// Subscribes to this module's events, filtering out those that do not pass [test].
  @optionalTypeArgs
  StreamSubscription<Event> where(String id, Value1Callback<Event, void> callback, ValueTest<Event> test);

  /// Subscribes to this module's events, filtering out those that are not of type [E].
  StreamSubscription<E> whereType<E extends Event>(String id, Value1Callback<E, void> callback);

  // static Module get root => modules[const []];
  /// The root module
  static RootModule get root => RootModule();

  /// The module registry
  static final Map<String, Module> modules = {
    // equals: const ListEquality<String>().equals,
    // hashCode: (k) => k.hashCode,
    // []: DefModule._([], Level.ALL, RootModule().eventHandler),
  };

  /// Returns the module with [package] if it exists, otherwise generates it with [builder]
  static Module getModuleSafely({@nullable List<String> package, @nullable String stringPackage, Level level, @required ModuleBuilder builder}) {
    if (!modules.containsKey(package?.join('.') ?? stringPackage)) modules[package?.join('.') ?? stringPackage] = builder(package ?? packageFromString(stringPackage), level);
    return modules[package?.join('.') ?? stringPackage];
  }

  /// Parses a [List<String>] package from [package].
  static List<String> packageFromString(String package) => package.split(RegExp(r'\.\/'));

  /// A function wrapping the StdModule constructor
  static StdModule generate(List<String> package, Level level) => StdModule._(package, level);
}

mixin HasModule implements Module {
  @override
  @protected
  List<String> get internalEventHandlers => module.internalEventHandlers;

  @override
  @protected
  List<String> get eventHandlers => module.eventHandlers;

  @protected
  Module get module => Module(package);

  @override
  String get immediateModule => module.immediateModule;

  @override
  Module get parent => module.parent;

  @override
  String get name => module.name;

  @override
  String get fullName => module.fullName;

  @override
  String get packageifiedPackage => module.packageifiedPackage;

  @override
  String get pathifiedPackage => module.pathifiedPackage;

  @override
  String get pathifiedPackageWithExtension => module.pathifiedPackageWithExtension;

  @override
  Level get recordStackTraceAtLevel => module.recordStackTraceAtLevel;

  @override
  @protected
  Level get internalLevel => module.internalLevel;

  @override
  @protected
  set internalLevel(Level newLevel) => module.internalLevel = newLevel;

  @override
  Level get level => module.level;

  @override
  set level(Level value) => module.level = value;

  @override
  bool isLoggable(Level value) => module.isLoggable(value);

  @override
  void log(Level logLevel, dynamic message, {@nullable String longMessage, @nullable Object error, @nullable StackTrace stackTrace, @nullable Zone zone, bool init = false, @nullable dynamic blob}) => module.log(logLevel, message, longMessage: longMessage, error: error, stackTrace: stackTrace, zone: zone, init: init, blob: blob);

  @override
  void logAuth(Level logLevel, dynamic message, AuthSignal signal, {@nullable String longMessage, @nullable Object error, @nullable StackTrace stackTrace, @nullable Zone zone, @nullable dynamic blob, bool init = false}) => module.logAuth(logLevel, message, signal, longMessage: longMessage, error: error, stackTrace: stackTrace, zone: zone, init: init, blob: blob);

  @override
  void logRequestInput(List<String> context, String prompt, InputTest test, {@nullable Object error, @nullable StackTrace stackTrace, @nullable Zone zone, @nullable dynamic blob, bool init = false}) => module.logRequestInput(context, prompt, test, error: error, stackTrace: stackTrace, zone: zone, init: init, blob: blob);

  @override
  StreamSubscription<Event> subscribe(String id, callback) => module.subscribe(id, callback);

  @override
  Future<void> unsubscribe(String id) => module.unsubscribe(id);

  @override
  Future<void> unsubscribeAll() => module.unsubscribeAll();

  @override
  StreamSubscription<Event> where(String id, callback, test) => module.where(id, callback, test);

  @override
  StreamSubscription<E> whereType<E extends Event>(String id, callback) => module.whereType<E>(id, callback);

  @override
  bool operator ==(other) => other is Module && const ListEquality<String>().equals(package, other.package);

  @override
  String toString() => pathifiedPackage;

  @override
  int get hashCode => package.hashCode;
}

mixin ModuleCommon on Module {
  @override
  final Level recordStackTraceAtLevel = Level.ERROR;

  @override
  String get immediateModule => package.isNotEmpty ? package.last : '/';

  @override
  Module get parent => Module(package.sublist(0, package.length - 1));

  @override
  String get name => immediateModule;

  /// The full name of this logger, which includes the parent's full name.
  @override
  String get fullName => packageifiedPackage;

  @override
  String get packageifiedPackage => '${package.isNotEmpty ? package.join('.') : 'ROOT'}';

  @override
  String get pathifiedPackage => '/${package.isNotEmpty ? package.join('/') : ''}';

  @override
  String get pathifiedPackageWithExtension => '/${package.isNotEmpty ? '${package.join('/')}.dart' : ''}';

  @override
  Level get level => internalLevel ?? parent.level ?? (internalLevel = RootModule.defaultLevel);

  @override
  set level(Level value) => (value != null || parent != null) ? internalLevel = value : internalLevel;

  @override
  bool isLoggable(Level value) => value >= level;

  @override
  void log(Level logLevel, dynamic message, {@nullable String longMessage, @nullable Object error, @nullable StackTrace stackTrace, @nullable Zone zone, bool init = false, dynamic blob}) {
    if (message is Function) {
      message = message();
    }

    if (message is! String) {
      blob ??= message;
      message = message.toString();
    }

    if (stackTrace == null && logLevel >= recordStackTraceAtLevel) {
      stackTrace = StackTrace.current;
      error ??= 'autogenerated stack trace for $logLevel $message';
    }

    zone ??= Zone.current;

    final event = StdEvent(this, logLevel, message, longMessage: longMessage, error: error, stackTrace: stackTrace, zone: zone, blob: blob, init: init);

    RootModule().add(event);
  }

  @override
  void logAuth(Level logLevel, dynamic message, AuthSignal signal, {@nullable String longMessage, @nullable Object error, @nullable StackTrace stackTrace, @nullable Zone zone, @nullable dynamic blob, bool init = false}) {
    if (message is Function) {
      message = message();
    }

    if (message is! String) {
      blob ??= message;
      message = message.toString();
    }

    if (stackTrace == null && logLevel >= recordStackTraceAtLevel) {
      stackTrace = StackTrace.current;
      error ??= 'autogenerated stack trace for $logLevel $message';
    }

    zone ??= Zone.current;

    final event = AuthEvent(this, logLevel, message, signal, longMessage: longMessage, error: error, stackTrace: stackTrace, zone: zone, blob: blob, init: init);

    RootModule().add(event);
  }

  @override
  void logRequestInput(List<String> context, String prompt, InputTest test, {@nullable Object error, @nullable StackTrace stackTrace, @nullable Zone zone, @nullable dynamic blob, bool init = false}) {
    zone ??= Zone.current;

    final event = RequestInputEvent(this, context, prompt, test, error: error, stackTrace: stackTrace, zone: zone, blob: blob, init: init);

    RootModule().add(event);
  }

  @override
  bool operator ==(other) => other is Module && packageifiedPackage == other.packageifiedPackage;

  @override
  String toString() => pathifiedPackage;

  @override
  int get hashCode => package.hashCode;
}

/// The standard [Module] implementation.
///
/// This should suit most use cases.
class StdModule extends Module with ModuleCommon {
  StdModule._(List<String> package, this.internalLevel)
      : assert(package.every((module) => !module.contains(RegExp(r'[^a-z0-9_]'))), 'Modules can only consist of lowercase alphanumeric characters and underscores.'),
        assert(package.every((module) => module.startsWith(RegExp(r'[a-z]'))), 'Modules must start with a lowercase letter.'),
        package = UnmodifiableListView(package),
        super.internal();

  @override
  final UnmodifiableListView<String> package;

  @override
  @nullable
  List<String> internalEventHandlers = [];

  @override
  @protected
  List<String> get eventHandlers => (internalEventHandlers ?? <String>[] + parent.eventHandlers).distinct().toList();

  @override
  Level internalLevel;

  @override
  StreamSubscription<Event> subscribe(String id, Value1Callback<Event, void> callback) => RootModule().where('$packageifiedPackage#$id', callback, (e) => e.source.eventHandlers.contains(id));

  @override
  Future<void> unsubscribe(String id) => RootModule().unsubscribe('$packageifiedPackage#$id');

  @override
  Future<void> unsubscribeAll() async {
    final root = RootModule();
    final _queue = Queue<MapEntry<String, StreamSubscription<Event>>>.of(root.subscriptions.entries.where((e) => e.key.startsWith(packageifiedPackage)));

    while (_queue.isNotEmpty) {
      final current = _queue.removeFirst();
      await root.subscriptions[current].cancel();
      root.subscriptions.remove(current);
    }
  }

  @override
  StreamSubscription<Event> where(String id, Value1Callback<Event, void> callback, ValueTest<Event> test) => RootModule().safeSubscribe(id, (_, queue) => queue.where((e) => e.source.eventHandlers.contains('$packageifiedPackage#$id')).where(test).listen((e) => callback(e)));

  @override
  StreamSubscription<E> whereType<E extends Event>(String id, Value1Callback<E, void> callback) => RootModule().safeSubscribe(id, (_, queue) => queue.where((e) => e.source.eventHandlers.contains('$packageifiedPackage#$id')).where((e) => e is E).cast<E>().listen((e) => callback(e)));
}
