import 'dart:async';
import 'dart:collection' show Queue;
import 'dart:io' as io;

import 'package:logging/logging.dart' as dlog;
import 'package:meta/meta.dart';
import 'package:flex/expose.dart';
import 'package:flex/flex.dart';
import 'package:plexus/plexus.dart';
import 'package:test_api/src/frontend/expect.dart' show TestFailure;

/// There is a single [RootModule] instance, which should **only** be managed (setting level, etc.) by the final product (i.e. Flutter app (not library) or CLI (seperated from api)) and should be accessed statically outside this library.
class RootModule extends Module with ModuleCommon {
  /// Returns the root module, generating it with `authEnabled`, `onInit`, `onDispose` and `level` if not found.
  factory RootModule({bool authEnabled = false, FutureOrCallback onInit = asyncVoidHandler, FutureOrCallback onDispose = asyncVoidHandler, Level level = defaultLevel}) => Module(const [], builder: (_, __) => RootModule._(PatientStreamQueue<Event>(), authEnabled, onInit, onDispose, level));

  RootModule._(this._internalQueue, this.authEnabled, this.onInit, this.onDispose, this.internalLevel) : super.internal();

  @override
  RootModule get parent => this;

  @override
  Level internalLevel;

  /// Whether authentication is enabled for the app
  final bool authEnabled;

  /// Called when [init] is called
  final FutureOrCallback<void> onInit;

  /// Called when [dispose] is called
  final FutureOrCallback<void> onDispose;

  /// The state of the app
  State state = State.uninitialized;

  @protected
  final PatientStreamQueue<Event> _internalQueue;

  @override
  Level get level => internalLevel ?? (level = defaultLevel);

  @override
  set level(Level value) {
    if (value != null) {
      internalLevel = value;
      dlog.Logger.root.level = value.dlogLevel;
    }
  }

  @override
  final UnmodifiableListView<String> package = const UnmodifiableListView([]);

  // void alert(dynamic error, [StackTrace stackTrace]) => tap.addError(error, stackTrace);

  /// The subscriptions registry.
  final Map<String, StreamSubscription<Event>> subscriptions = {};

  /// Safely subscribes to events
  ///
  /// If reference [id] is not already registered, registers the result of [generator].
  /// If specified, [E] can be used to tighten the return type, although casting of the stream itself must be handled by [generator].
  /// Keep in mind that if [E] is specified and [id] is already registered with a type other than [E], a TypeError will be thrown.
  @optionalTypeArgs
  StreamSubscription<E> safeSubscribe<E extends Event>(String id, StreamSubscription<E> Function(String id, PatientStreamQueue<Event> stream) generator) {
    if (subscriptions.containsKey(id)) {
      log(Level.WARNING, 'Subscription with id `$id` already exists. Not added.');
      try {
        return subscriptions[id];
      } on TypeError catch (error, stackTrace) {
        log(Level.ERROR, 'Subscription with id `$id` is a ${subscriptions[id].runtimeType}, not required type $E.', error: TypeError(), stackTrace: stackTrace);
        rethrow;
      }
    }

    return subscriptions[id] = generator(id, _internalQueue);
  }

  @override
  StreamSubscription<Event> subscribe(String id, Value1Callback<Event, void> callback) => safeSubscribe(id, (_, queue) => queue.listen(callback));

  @override
  Future<void> unsubscribe(String id) async {
    if (subscriptions.containsKey(id)) {
      await subscriptions[id].cancel();
      subscriptions.remove(id);
    } else {
      log(Level.WARNING, 'Subscription with id `$id` does not exist. Not removed.');
    }
  }

  @override
  Future<void> unsubscribeAll() async {
    final _queue = Queue<MapEntry<String, StreamSubscription<Event>>>.of(subscriptions.entries.where((e) => !e.key.startsWith('__') && !e.key.endsWith('__')));

    while (_queue.isNotEmpty) {
      final current = _queue.removeFirst();
      await subscriptions[current].cancel();
      subscriptions.remove(current);
    }
  }

  /// Returns a [StreamSubscription] of messages added to this [Logger] that pass [test].
  ///
  /// You can listen for messages using the standard stream APIs, for instance:
  ///
  /// ```dart
  /// logger.registerHandler((event) { ... });
  /// ```
  @override
  StreamSubscription<Event> where(String id, Value1Callback<Event, void> callback, ValueTest<Event> test) => safeSubscribe(id, (_, queue) => queue.where(test).listen((e) => callback(e)));

  @override
  StreamSubscription<E> whereType<E extends Event>(String id, Value1Callback<E, void> callback) => safeSubscribe(id, (_, queue) => queue.where((e) => e is E).cast<E>().listen((e) => callback(e)));

  /// Adds an event to the log
  void add(Event event) => _internalQueue.add(event);

  /// Adds an error event to the log
  void addError(Object error, [StackTrace stackTrace]) => _internalQueue.addError(error, stackTrace);

  Future<void> _registerInternalHandlers() async {
    if (authEnabled) whereType<AuthEvent>('__auth__', _authEventHandler);

    logAuth(Level.FINEST, 'Initial State', AuthSignal.SignedOut);
  }

  void _kernelEventHandler() {
    /// Terminal interrupt (usually Ctrl-C) -- Can be caught or ignored by any process | Useful for performing cleanup
    io.ProcessSignal.sigint.watch().listen((_) => dispose(wait: true, exitCode: 128 + 2));

    /// Terminal quit -- Sent by a terminal to indicate that the process should quit and perform a core dump | Can't be handled by Dart programs
    // io.ProcessSignal.sigquit.watch().listen((_) => dispose(wait: true, exitCode: 128 + 3));

    /// Kill -- can't be caught or ignored by any process (except the init process on *Nix-based systems)
    // io.ProcessSignal.sigkill.watch().listen((_) => dispose(wait: false, exitCode: 128 + 9));

    /// Requested termination -- Can be caught or ignored by any process | Useful for performing cleanup
    io.ProcessSignal.sigterm.watch().listen((_) => dispose(wait: false, exitCode: 128 + 15));
  }

  Future<void> _authEventHandler(AuthEvent event) async {
    @nullable
    Object _error;
    @nullable
    StackTrace _stackTrace;

    if (event.signal == AuthSignal.SignedOut || event.signal == AuthSignal.SignInFailed) {
      _stackTrace = event.stackTrace ?? StackTrace.current;
      _error = event.error; //?? 'Sorry, no error provided.';
    }

    switch (event.signal) {
      case AuthSignal.SignedOut:
        log(Level.ERROR, 'Sign In Failed.', longMessage: 'Sign in failed for some reason. This is the default SignInFailed handler. Maybe? try again.', error: _error, stackTrace: _stackTrace);
        break;
      case AuthSignal.SignInFailed:
        log(Level.ERROR, 'Sign In Failed.', longMessage: 'Sign in failed for some reason. This is the default SignInFailed handler. Maybe? try again.', error: _error, stackTrace: _stackTrace);
        break;
      case AuthSignal.SignedIn:
        log(Level.SUCCESS, 'We\'re Signed In!');
        break;
      case AuthSignal.SignOutFailed:
        break;
    }
  }

  /// Start the app.
  ///
  /// **MUST be called by the end user/app. NOT by libraries/packages**
  ///
  /// Should only be used once in the lifecycle of the app. Using more than once is not tested and is likely to fail.
  Future<void> start(FutureOr<void> main(), {Map<String, EventHandler> initialHandlers = const {}}) async {
    if (state == State.uninitialized) {
      await runZoned<Future<void>>(
        () async {
          state = State.initializing;

          try {
            _kernelEventHandler();
          } catch (error, stackTrace) {
            log(Level.ERROR, 'Error occurred during registration of internal handlers.', error: error, stackTrace: stackTrace, zone: Zone.current);
            rethrow;
          }

          // try {
          log(Level.FINER, 'Initializing Logger...', init: true);
          level = defaultLevel;
          dlog.Logger.root.onRecord.listen((record) {
            final shortMessage = record.message.length > 80 ? '${record.message.substring(0, 80 - 3)}...' : record.message;
            final longMessage = record.message.length > 80 ? record.message : null;
            add(StdEvent(
              Module(record.loggerName.toLowerCase().replaceAll(r'[^a-z0-9_\.]', '_').split('.')),
              Level.dlogToPlexus[record.level],
              shortMessage,
              longMessage: longMessage,
              error: record.error,
              stackTrace: record.stackTrace,
              blob: record.object,
              timestamp: record.time,
              zone: record.zone,
            ));
          });
          log(Level.FINE, 'Initialized logger', init: true);
          // } catch (error, stackTrace) {
          //   // log(Level.ERROR, 'Error occurred during initialization of logger.', error: error, stackTrace: stackTrace, zone: Zone.current);
          //   rethrow;
          // }

          try {
            await _registerInternalHandlers();
          } catch (error, stackTrace) {
            log(Level.ERROR, 'Error occurred during registration of internal handlers.', error: error, stackTrace: stackTrace, zone: Zone.current);
            rethrow;
          }

          try {
            initialHandlers.forEach((id, handler) => subscribe(id, handler));
          } catch (error, stackTrace) {
            log(Level.ERROR, 'Error occurred during registration of initial handlers.', error: error, stackTrace: stackTrace, zone: Zone.current);
            rethrow;
          }

          try {
            log(Level.FINER, 'Initializing Plexus Root Module...', init: true);

            await onInit();

            log(Level.FINE, 'Initialized Plexus Root Module.', init: true);
          } catch (error, stackTrace) {
            log(Level.ERROR, 'Error occurred during initialization of Plexus root module.', error: error, stackTrace: stackTrace, zone: Zone.current);
            rethrow;
          }

          state = State.initialized;

          try {
            await main();
          } catch (error, stackTrace) {
            if (error is TestFailure) rethrow;

            log(Level.ERROR, 'Error occurred during execution of main().', error: error, stackTrace: stackTrace, zone: Zone.current);
          }

          await dispose();
        },
        zoneSpecification: ZoneSpecification(
          print: (_, __, zone, message) {
            if (message.startsWith(encodedEventHeader)) {
              final event = Event.parse(message, this, zone);
              _internalQueue.isClosed ? io.stdout.writeln('[${event.level}]: ${event.message}') : add(event);
            } else {
              _internalQueue.isClosed ? io.stdout.writeln(message) : log(Level.PRINT, message, zone: zone);
            }
          },
          handleUncaughtError: (_, __, zone, error, stackTrace) {
            // try {
            log(Level.FATAL, 'Error occurred within Plexus root module isolation zone that was not handled. This may be due to an error in initialization of the logger.', error: error, stackTrace: stackTrace, zone: zone);
            // } catch (lError, logStackTrace) {
            //   rethrow;
            // }

            // io.stdout.writeln('Error occurred within Plexus root module isolation zone that was not handled. This should not be possible. $error | $stackTrace');
            dispose();
          },
        ),
      );
    }
  }

  /// Disposes of the root module.
  ///
  /// **MUST be called by the end user/app. NOT by libraries/packages**
  ///
  /// Should only be used once in the lifecycle of the app. Using more than once is not tested and is likely to fail.
  Future<void> dispose({bool wait = true, int exitCode = 0}) async {
    print('Termination in progress...');

    await onDispose();

    if (wait) {
      await _internalQueue.close(wait: wait);
    } else {
      _internalQueue.dispose();
    }

    print('Goodbye.');

    io.exit(exitCode);
  }

  @override
  final List<String> internalEventHandlers = [];

  @override
  @protected
  List<String> get eventHandlers => internalEventHandlers;

  /// The default level for [RootModule]
  static const Level defaultLevel = Level.FINE;
}
